-- Temporary hack to allow type checking of simulator code without having
-- to build the simulator.
-- Import this one instead of AFrob.RobotSim.

module AFrob.RobotSim.Fake (
    SimbotProperties,
    SimbotInput,
    SimbotOutput,
    SimbotController,
    WorldTemplate,
    ObjectTemplate(..),
    BoundingBox,
    worldXMin,
    worldYMin,
    worldXMax,
    worldYMax,
    boundingBox,
    runSim
) where

import AFRP
import AFRP.MergeableRecord (MergeableRecord(..), MR, mrMake)
import AFrob.PhysicalDimensions
import AFrob.RobotIO

------------------------------------------------------------------------------
-- Property type and instances
------------------------------------------------------------------------------

data SimbotProperties = SP {
    -- Fields for HasRobotProperties
    spRType    :: RobotType,	-- "SimbotA" and "SimbotB"
    spRId      :: RobotId,
    spDiameter :: Length,
    spAccMax   :: Acceleration,
    spWSMax    :: Speed
}


instance HasRobotProperties SimbotProperties where
    rpType     = spRType
    rpId       = spRId
    rpDiameter = spDiameter
    rpAccMax   = spAccMax
    rpWSMax    = spWSMax


------------------------------------------------------------------------------
-- Input type and instances
------------------------------------------------------------------------------

-- !!! Should the fields be strict?

-- !!! Should some of the fields, like the field for stuck, really be
--     an event?

data SimbotInput = SI {
    -- Fields for HasRobotStatus
    siBattStat :: BatteryStatus,
    siIsStuck  :: Bool,		-- Really an event?

    -- Fields for HasOdometry
    siPosition :: Position2,
    siHeading  :: Heading,

    -- Fields for HasRangeFinder
    siRanges   :: [Distance],		-- n equispaced range finders, CCW.
    siMaxRange :: Distance
}


instance HasRobotStatus SimbotInput where
    rsBattStat = siBattStat
    rsIsStuck  = siIsStuck


instance HasOdometry SimbotInput where
    odometryPosition = siPosition
    odometryHeading  = siHeading


instance HasRangeFinder SimbotInput where
    rfRange i a = rs !! (round (fromIntegral n * a / (2 * pi)) `mod` n)
        where
	    rs = siRanges i
	    n = length rs

    rfMaxRange = siMaxRange


instance HasAnimateObjectTracker SimbotInput where
    aotOtherRobots = undefined
    aotBalls       = undefined


------------------------------------------------------------------------------
-- Output type with instances and subordinate types
------------------------------------------------------------------------------

data DiffDriveMode =
      DDMBrake				-- Brake both wheels.
    | DDMDiff {
	  ddmdLWV :: Velocity,		-- Left Wheel Velocity.
	  ddmdRWV :: Velocity		-- Right Wheel Velocity.
      }
    | DDMTR {
	  ddmtrTV :: Velocity,		-- Translational Velocity.
	  ddmtrRV :: RotVel		-- Rotational Velocity.
      }

-- !!! Should the text console output be an event? Yes probably.

data SimbotOutput = SO {
    soDD   :: DiffDriveMode,		-- Differential drive output.
    soTCO  :: Maybe String		-- Text console output.
--    soFVGO :: SimpleGraphic		-- FVision GUI output.
}


instance MergeableRecord SimbotOutput where
    mrDefault = SO {soDD = DDMBrake, soTCO = Nothing}


instance HasDiffDrive SimbotOutput where
    ddBrake = mrMake (\o -> o {soDD = DDMBrake})

    ddVelDiff lwv rwv =
        mrMake (\o -> o {soDD = DDMDiff {ddmdLWV = lwv, ddmdRWV = rwv}})

    ddVelTR v rv =
        mrMake (\o -> o {soDD = DDMTR {ddmtrTV = v, ddmtrRV = rv}})


instance HasTextConsoleOutput SimbotOutput where
    tcoPrintMessage em = undefined
{-
        mrMake (\o -> event o (\m -> o {soTCO = m : soTCO o}) em)
-}

------------------------------------------------------------------------------
-- Simbot controller
------------------------------------------------------------------------------

type SimbotController = SimbotProperties -> SF SimbotInput SimbotOutput


------------------------------------------------------------------------------
-- World and object templates
------------------------------------------------------------------------------

type WorldTemplate = [ObjectTemplate]


data ObjectTemplate =
      OTBlock {			-- Square-shaped obstacle
	  otPos  :: Position2
      }
    | OTNSWall {		-- North-South wall segment
	  otPos  :: Position2
      }
    | OTEWWall {		-- East-west wall segment
	  otPos  :: Position2
      }
    | OTVWall {			-- "Vertical" wall segment, same as North-South
	  otPos  :: Position2
      }
    | OTHWall {			-- "Horizontal" wall segment, same as east-west
	  otPos  :: Position2
      }
    | OTSimbotA {		-- Robot of type Simbot A
          otRId  :: RobotId,
	  otPos  :: Position2,
	  otHdng :: Heading
      }
    | OTSimbotB {		-- Robot of type Simbot B
          otRId  :: RobotId,
	  otPos  :: Position2,
	  otHdng :: Heading
      }
    | OTBall {			-- Ball
	  otPos  :: Position2
      }


type BoundingBox = (Position2, Position2)

-- This is probably not such a great way of finding out the size of objects?
boundingBox :: ObjectTemplate -> BoundingBox
boundingBox = undefined


------------------------------------------------------------------------------
-- World-related defs.
------------------------------------------------------------------------------

-- The world is assumed to be rectangular.

worldXMin, worldXMax :: Position
worldYMin, worldYMax :: Position
worldXMin = -5.0
worldYMin = -5.0
worldXMax = 5.0
worldYMax = 5.0


------------------------------------------------------------------------------
-- Main interface
------------------------------------------------------------------------------

runSim :: Maybe WorldTemplate -> SimbotController -> SimbotController -> IO ()
runSim _ _ _ = undefined
